#ifdef MAINTAINER_MODE
# include <sys/types.h>
# include <sys/wait.h>
# include <unistd.h>
# include <stdio.h>
# include <errno.h>
#endif
#include <matlabint_misc.h>
#include <getfem_mesh_fem.h>
#include <algorithm>

namespace matlabint {
  gfi_array *
  convert_to_gfi_sparse(const gf_sparse_by_row& smat, double threshold)
  {
    int ni = gmm::mat_nrows(smat);
    int nj = gmm::mat_ncols(smat);
    unsigned nnz = 0;
    std::vector<int> ccnt(nj);
    gfi_array *mxA;
  
  
    std::fill(ccnt.begin(), ccnt.end(), 0);

    std::vector<double> rowmax(ni), colmax(nj);

    std::fill(rowmax.begin(), rowmax.end(), 0);
    std::fill(colmax.begin(), colmax.end(), 0);

    gmm::linalg_traits<gmm::linalg_traits<gf_sparse_by_row>
      ::const_sub_row_type>::const_iterator it, ite;
    
    /* first pass : find the maxima / row & column */
    for (int i = 0; i < ni; ++i) {
      it = gmm::vect_const_begin(gmm::mat_const_row(smat,i)); 
      ite = gmm::vect_const_end(gmm::mat_const_row(smat,i));
      for (; it != ite; ++it) {
	rowmax[i] = std::max(rowmax[i], dal::abs(*it));
	colmax[it.index()] = std::max(colmax[it.index()], dal::abs(*it));
      }
    }

    /* second pass : count the significant terms */
    for (int i = 0; i < ni; i++) {
      /* count significatives values */
      it = gmm::vect_const_begin(gmm::mat_const_row(smat,i)); 
      ite = gmm::vect_const_end(gmm::mat_const_row(smat,i));
      for (; it != ite; ++it) {
	if (*it != 0. && /* (*p).e != 0 => max(rowmax,colmax) != 0 */
	    dal::abs(*it) > threshold * std::max(rowmax[i], colmax[it.index()]) ) {
	  ccnt[it.index()]++;
	  nnz++;
	}
      }
    }

    mxA = gfi_create_sparse(ni, nj, nnz); assert(mxA != NULL);
    double *pr;
    unsigned *ir, *jc;
    pr = gfi_sparse_get_pr(mxA); assert(pr != NULL);
    ir = gfi_sparse_get_ir(mxA); assert(ir != NULL);
    jc = gfi_sparse_get_jc(mxA); assert(jc != NULL); /* dim == nj+1 */
  
    jc[0] = 0;
    for (int j=0; j < nj; j++) {
      //    mexPrintf("ccnt[%d]=%d\n", j, ccnt[j]);
      jc[j+1] = jc[j] + ccnt[j];
    }
    assert(nnz == jc[nj]);

    //gmm::wsvector<getfem::scalar_type>::const_sorted_iterator its, itse;
    gmm::linalg_traits<gmm::rsvector<getfem::scalar_type> >::const_iterator its, itse;

    /* third pass: filling pr and ir */
    std::fill(ccnt.begin(), ccnt.end(), 0);
    gmm::rsvector<getfem::scalar_type> sorted(gmm::mat_ncols(smat));
    for (int i=0; i < ni; i++) {
      gmm::copy(gmm::mat_const_row(smat, i), sorted);
      its  = gmm::vect_begin(sorted);
      itse = gmm::vect_end(sorted);
      for (; its != itse; ++its) {
	if ((*its) != 0. && /* (*its) != 0 => max(rowmax,colmax) != 0 */
	    dal::abs(*its)/std::max(rowmax[i], colmax[its.index()]) > threshold) {
	  ir[jc[its.index()]+ccnt[its.index()]] = i;
	  pr[jc[its.index()]+ccnt[its.index()]] = *its;
	  ccnt[its.index()]++;
	}
      }
    }
    return mxA;
  }

  gfi_array *
  convert_to_gfi_sparse(const gf_sparse_by_col& smat, double threshold)
  {
    
    int ni = gmm::mat_nrows(smat);
    int nj = gmm::mat_ncols(smat);
    unsigned nnz = 0;
    std::vector<int> ccnt(nj);
    gfi_array *mxA;

    std::fill(ccnt.begin(), ccnt.end(), 0);

    std::vector<double> rowmax(ni), colmax(nj);
    std::fill(rowmax.begin(), rowmax.end(), 0);
    std::fill(colmax.begin(), colmax.end(), 0);

    gmm::linalg_traits<gmm::linalg_traits<gf_sparse_by_col>
      ::const_sub_col_type>::const_iterator it, ite;

    /* first pass : find the maxima / row & column */
    for (int j = 0; j < nj; ++j) {
      it = gmm::vect_const_begin(gmm::mat_const_col(smat,j)); 
      ite = gmm::vect_const_end(gmm::mat_const_col(smat,j));
      for (; it != ite; ++it) {
	rowmax[it.index()] = std::max(rowmax[it.index()], dal::abs(*it));
	colmax[j] = std::max(colmax[j], dal::abs(*it));
      }
    }

   /* second pass : count the significant terms */
    for (int j = 0; j < nj; j++) {
      /* count significatives values */
      it = gmm::vect_const_begin(gmm::mat_const_col(smat,j)); 
      ite = gmm::vect_const_end(gmm::mat_const_col(smat,j));
      for (; it != ite; ++it) {
	if (*it != 0. && /* (*p).e != 0 => max(rowmax,colmax) != 0 */
	    dal::abs(*it) > threshold * std::max(colmax[j], rowmax[it.index()]) ) {
	  ccnt[j]++;
	  nnz++;
	}
      }
    }

    mxA = gfi_create_sparse(ni, nj, nnz); assert(mxA != NULL);
    double *pr;
    unsigned *ir, *jc;
    pr = gfi_sparse_get_pr(mxA); assert(pr != NULL);
    ir = gfi_sparse_get_ir(mxA); assert(ir != NULL);
    jc = gfi_sparse_get_jc(mxA); assert(jc != NULL); /* dim == nj+1 */
  
    jc[0] = 0;
    for (int j=0; j < nj; j++) {
      //    mexPrintf("ccnt[%d]=%d\n", j, ccnt[j]);
      jc[j+1] = jc[j] + ccnt[j];
    }
    assert(nnz == jc[nj]);

    gmm::linalg_traits<gmm::rsvector<getfem::scalar_type> >::const_iterator its, itse;

    /* third pass: filling pr and ir */
    std::fill(ccnt.begin(), ccnt.end(), 0);

    gmm::rsvector<getfem::scalar_type> sorted(gmm::mat_nrows(smat));
    for (int j=0; j < nj; j++) {
      gmm::copy(gmm::mat_const_col(smat, j), sorted);
      its  = gmm::vect_const_begin(sorted);
      itse = gmm::vect_const_end(sorted);
      for (; its != itse; ++its) {
	if ((*its) != 0. && /* (*its) != 0 => max(rowmax,colmax) != 0 */
	    dal::abs((*its))/std::max(colmax[j], rowmax[its.index()]) > threshold) {
	  ir[jc[j]+ccnt[j]] = its.index();
	  pr[jc[j]+ccnt[j]] = (*its);
	  ccnt[j]++;
	}
      }
    }
    return mxA;
  }


  /* merge the edge list of a convex into 'el', 
     filtering the edges of the face 'f' if (f != -1) */
  static
  void mesh_edge_list_merge(const getfem::getfem_mesh &m, bgeot::edge_list &el, const bgeot::edge_list &elcv, int cv, int f)
  {
    bgeot::edge_list::const_iterator it = elcv.begin();
    for (it = elcv.begin(); it != elcv.end(); it++) {
      if (f != -1) {
	bgeot::ind_ref_mesh_point_ind_ct pt = m.ind_points_of_face_of_convex(cv, f);
	if (std::find(pt.begin(), pt.end(), (*it).i) == pt.end()) continue;
	if (std::find(pt.begin(), pt.end(), (*it).j) == pt.end()) continue;
      }
      el.add(*it);
    }
  }

  /* used by gf_mesh_get and gf_mesh_fem_get */
  void
  build_edge_list(const getfem::getfem_mesh &m, bgeot::edge_list &el, mexargs_in &in) {
    mlab_vect v;
    bool all_cv = true;
    bool merge_convex = false;
    
    if (in.remaining() && !in.front().is_string()) {
      v = in.pop().to_int_vector(-1, -1);
      all_cv = false;
    }
    if (in.remaining() && in.front().is_string()) {
      std::string s = in.pop().to_string();
      if (cmd_strmatch(s,"merge convex")) 
	merge_convex = true;
      else bad_cmd(s);
    }
    if (all_cv) { /* the fastest way : edges off all convexes */
      bgeot::mesh_edge_list(m, el, merge_convex);
    } else {              /* slower (and not very smart) way : edges from 
			     selected convexes/faces */
      if (v.getm() != 1 && v.getm() != 2) 
	THROW_ERROR("wrong number of rows for CVLIST");
      
      size_type j = 0;
      /* loop over the rows -- if face numbers are given, the performance will
	 be better if the columns are grouped by convex numbers */
      while (j < v.getn()) {
	size_type cv = size_type(v(0,j)-1);
	bgeot::edge_list elcv;

	if (!m.convex_index().is_in(cv)) 
	  THROW_ERROR("can't build edges of convex " << cv 
		       << ": there is no such convex in mesh");
        std::vector<size_type> cvpt(m.nb_points_of_convex(cv));
        std::copy(m.ind_points_of_convex(cv).begin(), m.ind_points_of_convex(cv).end(), cvpt.begin());
	bgeot::mesh_edge_list_convex(m.structure_of_convex(cv), cvpt, cv, elcv, merge_convex);	
	if (v.getm() == 2) { /* face numbers present */
	  /* loop while the convex number is do not change */
	  do {
	    mesh_edge_list_merge(m, el, elcv, cv, size_type(v(1,j)-1));
	    j++;
	  } while (j < v.getn() && size_type(v(0,j)-1) == cv);
	} else { /* merge all edges of the convex */
	  mesh_edge_list_merge(m,el, elcv, cv, -1);
	  j++;
	}
      }
    }
  }

  /* build a vector of couples <convex number, face number> */
  void build_convex_face_lst(const getfem::getfem_mesh& m, std::vector<convex_face>& l, const mlab_vect *v) {
    l.resize(0);
    if (v) {
      if (v->getm() != 1 && v->getm()!=2) THROW_ERROR("too much rows (2 max)");
      l.resize(v->getn());
      for (size_type i=0; i < v->getn(); ++i) {
	l[i].cv = size_type(v->operator()(0,i))-1;
	if (!m.convex_index()[l[i].cv]) 
	  THROW_ERROR("the convex " << l[i].cv+1 << " is not part of the mesh");
	if (v->getm() == 2) {
	  l[i].f  = dim_type(v->operator()(1,i)-1);
	  if (dim_type(l[i].f+1) && l[i].f >= 
	      m.structure_of_convex(l[i].cv)->nb_faces())
	    THROW_ERROR("face " << l[i].f+1 << " of convex " << l[i].cv+1 << "(" 
			<< bgeot::name_of_geometric_trans(m.trans_of_convex(l[i].cv))
			<< ") does not exist");
	}
	else l[i].f = dim_type(-1);
      }
    } else {
      l.reserve(m.convex_index().card());
      for (dal::bit_vector::const_iterator it = m.convex_index().begin();
	   it != m.convex_index().end(); ++it) {
	if (*it) l.push_back(convex_face(it.index(),dim_type(-1)));
      }
    }
  }

  /*
    interpolate the solution (in vector U) on the element cv,
    evaluated at the points 'pt' (given in the reference convex)
  */
  void interpolate_on_convex_ref(const getfem::mesh_fem *mf, getfem::size_type cv, 
				 const bgeot::vsvector<getfem::base_node> &pt, 
				 const mlab_vect& U,
				 getfem::base_matrix &pt_val)
  {
    assert(mf->convex_index().is_in(cv));
    getfem::pfem cv_fem = mf->fem_of_element(cv); 
    size_type qdim = mf->get_qdim();
    /* largely inspired by getfem_export.h */

    /* interpolation of the solution.                                  */
    /* faux dans le cas des lments non tau-equivalents ou vectoriel. */
    if (cv_fem->target_dim() != 1) 
      THROW_ERROR("interpolation on vector fem is still to be done! (or at least to be tested...)");
    if (U.getn() != mf->nb_dof())
      THROW_ERROR("wrong nb of columns for U");
    assert(cv_fem->is_equivalent());
  
    pt_val.resize(qdim*U.getm(), pt.size());

    getfem::base_matrix G;

    if (mf->fem_of_element(cv)->need_G()) 
      getfem::transfert_to_G(G, mf->linked_mesh().points_of_convex(cv));

    getfem::base_vector coeff(cv_fem->nb_dof());

    getfem::base_node val(cv_fem->target_dim());

    for (size_type row = 0; row < U.getm(); ++row) {
      if (cv_fem->target_dim() == 1) {
	for (size_type dim = 0; dim < qdim; dim++) {
	  for (size_type j = 0; j < coeff.size(); j++) {
	    coeff[j] = U[mf->ind_dof_of_element(cv)[j*qdim + dim]];
	  }  
	  for (size_type j = 0; j < pt.size(); j++) {
	    cv_fem->interpolation(pt[j], G, mf->linked_mesh().trans_of_convex(cv), coeff, val);
	    pt_val(row*qdim + dim, j) = val[0];
	  }
	}
      } else { /* vector elements */
	for (size_type j = 0; j < coeff.size(); j++) {
	  coeff[j] = U[mf->ind_dof_of_element(cv)[j]];
	}  
	for (size_type j = 0; j < pt.size(); j++) {
	  cv_fem->interpolation(pt[j], G, mf->linked_mesh().trans_of_convex(cv), coeff, val);
	  for (size_type q = 0; q < qdim; ++q)
	    pt_val(row*qdim + q,j) = val[j];
	}
      }
    }
  }

  /* utility function for eval_on_triangulated_surface */
  static void 
  eval_sub_nodes(unsigned N, const std::vector<getfem::base_node>& V, std::vector<getfem::base_node>& spt)
  {
    assert(N>0);
    spt.resize(((N+1)*(N+2))/2);
    
    /*
      .      pt[0]
      .        /\
      .       /  \
      . pt[2]/____\pt[1]
    
      refinment:
    
      .         0     <- layer 0
      .        111    <- layer 1 etc..
      .       22222
    
    */
    
    spt[0] = V[0];
    size_type pcnt = 1;

    /* find the three nodes of the each sub-triangle */
    for (size_type layer = 1; layer <= N; layer++) {
      getfem::base_node A,B;
      getfem::scalar_type c;
      c = ((getfem::scalar_type)layer)/N;
    
      A = V[0] + (V[2] - V[0]) * c;
      B = V[0] + (V[1] - V[0]) * c;

      for (size_type inode = 0; inode <= layer; inode++, pcnt++) {
	spt[pcnt] = A + (B-A) * ((inode)/(getfem::scalar_type)layer);
      }
    }
    if (!(pcnt == spt.size())) THROW_INTERNAL_ERROR;
  }

  /* U may have zero rows, in that case, only the geometric deformation
     of the points will be computed and stored in w
  */
  static void
  add_refined_tri(const getfem::getfem_mesh *mesh, size_type cv, 		  
		  const std::vector<getfem::base_node>& vertices, int Nrefine, 
		  mlab_vect& w, size_type tri_cnt, 
		  const getfem::mesh_fem *pmf, const mlab_vect& U)
  {
    unsigned mesh_dim = mesh->dim();
    unsigned qdim = (pmf) ? pmf->get_qdim() : 0;
    bgeot::pgeometric_trans pgt = mesh->trans_of_convex(cv);
    bgeot::vsvector<getfem::base_node> pt(Nrefine * Nrefine);

    eval_sub_nodes(Nrefine, vertices, pt);

    getfem::base_matrix pt_val;
    if (pmf) {
      check_cv_fem(*pmf, cv);
      pt_val.resize(qdim*U.getm(), pt.size());
      interpolate_on_convex_ref(pmf, cv, pt, U, pt_val);
    }

    /* apply the geometric transformation to the points, in order to
       find their real location on the mesh */
    //getfem::base_node P(mesh_dim); 
    for (size_type i=0; i < pt.size(); ++i) {
      pt[i] = pgt->transform(pt[i], mesh->points_of_convex(cv));
      /*      P.fill(0.0);
      for (getfem::size_type j = 0; j < pgt->nb_points(); ++j) {
	P.addmul(pgt->poly_vector()[j].eval(pt[i].begin()),
		 mesh->points_of_convex(cv)[j]);
		 }*/
      //pt[i] = P;
    }

    size_type refined_tri_cnt = 0;
    /* find the three nodes of the each sub-triangle */
    for (int layer = 0; layer < Nrefine; layer++) {
      for (int itri = 0; itri < layer*2+1; itri++, refined_tri_cnt++) {
	getfem::size_type n[3];

	if ((itri & 1) == 0) {
	  /*
	    .           0
	    .          /\
	    .       2 /__\ 1
	  */
	  n[0] = (layer*(layer+1))/2 + itri/2;
	  n[1] = n[0] + layer+1; 
	  n[2] = n[1]+1;
	} else {
	  /*
	    .       1 ____ 2
	    .         \  / 
	    .          \/
	    .           0
	  */
	  n[1] = (layer*(layer+1))/2 + itri/2;
	  n[2] = n[1]+1;
	  n[0] = n[2]+layer+1;
	}
	
	if (!(n[0] < pt.size()) || !(n[1] < pt.size()) || !(n[2] < pt.size())) {
	  THROW_INTERNAL_ERROR;
	}	
	for (size_type ipt = 0; ipt < 3; ++ipt) {
	  for (size_type idim = 0; idim < mesh_dim; ++idim) {
	    w(ipt * mesh_dim + idim, tri_cnt+refined_tri_cnt) = pt[n[ipt]][idim];
	  }
	  if (pmf) {
	    for (size_type row = 0; row < U.getm(); ++row) {
	      for (size_type q = 0; q < qdim; ++q) {
		w(3*mesh_dim + row*qdim*3 + ipt*qdim + q,  tri_cnt + refined_tri_cnt) = pt_val(q,n[ipt]);
	      }
	    }
	  }
	}
      }
    }
  }

/* 
   interpolate a field U defined on mf onto a P1 discontiuous mesh should work
   for planar segment/triangle/quadrangle meshes, and faces of volumic meshes
   which are a segment/triangle/quadrangle. In order to be able to represent
   accurately high order FEMs, a refinment factor can be specified.

   the aim is to be fast (the matlab graphic routines don't need any additional
   slowness...)

   This function can still be used if pmf is NULL, in order to obtain a nice
   decomposition of a mesh into triangle (interesting for non-linear geometric
   deformations, used by gf_plot_mesh)
*/  void
  eval_on_triangulated_surface(const getfem::getfem_mesh* mesh, int Nrefine,
			       const std::vector<convex_face>& cvf,
			       mexargs_out& out,
			       const getfem::mesh_fem *pmf, const mlab_vect& U) {
    unsigned mesh_dim = mesh->dim();
    unsigned qdim = (pmf ? pmf->get_qdim() : 0);
    if (mesh_dim < 2 || mesh_dim > 3) {
      THROW_ERROR( "This function do not handle " << 
		   mesh_dim << "D meshes (only 2D or 3D)");
    }
  
    /* first pass: count nb of triangles */
    size_type nb_tri = 0;
    for (size_type i=0; i < cvf.size(); ++i) {
      bgeot::pconvex_ref cv_ref=mesh->trans_of_convex(cvf[i].cv)->convex_ref();

      bgeot::pconvex_structure cv_struc=cv_ref->structure();

      if (cvf[i].f != dim_type(-1)) {
	cv_struc = cv_struc->faces_structure()[cvf[i].f];
      }
      if (cv_struc->basic_structure()->nb_points() == 2) continue; /* pas les lignes */
      if (cv_struc->dim() > 2) 
	THROW_ERROR("cannot draw a 3D convex (convex nb " << cvf[i].cv << "), please specify "
		    "faces (see gf_mesh_get(m,'outer faces') for example)");
      size_type t_inc = 0;
      switch (cv_struc->basic_structure()->nb_points()) {
      case 3: t_inc = 1; break;
      case 4: t_inc =2; break;
      }
      if (pmf == NULL && mesh->trans_of_convex(cvf[i].cv)->is_linear()) {
	nb_tri += t_inc;
      } else {
	nb_tri += Nrefine*Nrefine*t_inc;
      }
    }

    mlab_vect w = out.pop().create_vector(3*mesh_dim + 3*qdim*U.getm(), nb_tri);

    /* second pass */
    size_type tri_cnt = 0;
    std::vector<size_type> ipts;
    for (size_type i=0; i < cvf.size(); ++i) {
      std::vector<getfem::base_node> vertices(3);
      size_type n_refine;
      bgeot::pconvex_ref cv_ref=mesh->trans_of_convex(cvf[i].cv)->convex_ref();
      bgeot::pconvex_structure cv_struc=cv_ref->structure();

      if (pmf == NULL && mesh->trans_of_convex(cvf[i].cv)->is_linear())
	n_refine = 1;
      else n_refine = Nrefine;
      
      if (cvf[i].f == dim_type(-1)) {
	ipts.resize(cv_ref->nb_points());
	for (size_type j=0; j < ipts.size(); ++j) ipts[j]=j;
      } else {
	const bgeot::convex_ind_ct& pts = cv_struc->ind_points_of_face(cvf[i].f);
	ipts.resize(pts.size());
	std::copy(pts.begin(), pts.end(), ipts.begin());
	cv_struc = cv_struc->faces_structure()[cvf[i].f];
      }
      if (cv_struc->basic_structure()->nb_points() == 2) continue; /* pas les lignes */

      /* for each point, count the nb of faces it belongs */
      std::vector<size_type> fcnt(cv_struc->nb_points(), 0);
      for (size_type f = 0; f < cv_struc->nb_faces(); ++f) {
	for (size_type fp = 0; fp < cv_struc->nb_points_of_face(f); ++fp) {
	  fcnt[cv_struc->ind_points_of_face(f)[fp]]++;
	}
      }
      /* remove unused points */
      size_type j=0;
      for (size_type k=0; k < ipts.size(); ++k) {
	if (fcnt[k]==2) {
	  ipts[j] = ipts[k];
	  j++;
	}
      }
      ipts.resize(j);      

#define DO_TRIANGLE(a,b,c) { \
      vertices[0] = cv_ref->points()[ipts[a]]; \
      vertices[1] = cv_ref->points()[ipts[b]]; \
      vertices[2] = cv_ref->points()[ipts[c]]; \
      add_refined_tri(mesh, cvf[i].cv, vertices, \
                      n_refine, \
                      w, tri_cnt, pmf, U); tri_cnt += n_refine*n_refine; }

      if (ipts.size() == 3) {
	DO_TRIANGLE(0,1,2);
      } else if (ipts.size() == 4) {
	DO_TRIANGLE(0,1,2); 
	DO_TRIANGLE(1,3,2);
      } else {
	cerr << "convex not handled by eval_on_triangulated_surface: " 
	     << bgeot::name_of_geometric_trans(mesh->trans_of_convex(cvf[i].cv)) << endl;
	//cerr << "cv_struc->basic_structure()->nb_points() = " << cv_struc->basic_structure()->nb_points() << endl;
	//	DAL_INTERNAL_ERROR("");
      }
      assert(tri_cnt <= nb_tri);
    }
    if (tri_cnt != nb_tri) DAL_INTERNAL_ERROR("tri_cnt=" << tri_cnt << ", nb_tri=" << nb_tri);
  }


  /* because it's such a pain to launch matlab with a debugger.. */
  void attach_gdb() {
#   ifdef MAINTAINER_MODE
    pid_t pid, parent;
    char *cmd;
    char cmd_pid[1024];
    cmd = ::getenv("GETFEM_DEBUG_CMD"); /* i.e. "xterm -e gdb /path/to/bin/matlab %d" 
					   (xterm because matlab does weird things with its tty (no echo of input..))
					*/
    if (cmd == NULL) {
      cerr << "the environment variable $GETFEM_DEBUG_CMD is not set, cancelling the debugger\n"
	   << "You should set it to something like \"xterm -e gdb /path/to/bin/matlab %d\" (%d will be replaced by the matlab pid)\n";
      return;
    }
    parent = ::getpid();
    ::snprintf(cmd_pid, 1024, cmd, parent);
    std::cerr << "internal error detected\nrunning /bin/sh -c \"" << cmd_pid << "\" ...\n" << std::endl;
    switch ((pid=::fork())) {
    case -1: std::cerr << "can't fork gdb\n"; return;
    case 0: /* child */
      {
	execlp("/bin/sh", "/bin/sh", "-c", cmd_pid, NULL);
	cerr << "execlp failure .. : " << ::strerror(errno) << endl;
	exit(1); /* in case the exec failed */
      } break;
    default:
      {
	int status;
	std::cout << "When you have finished with gdb, just quit it (type q)\n";
	::waitpid(pid, &status, 0);
	std::cout << "gdb done!\n";
      } break;
    }
#   endif /*MAINTAINER_MODE*/
  }
}

